/*
 *
 *   Copyright (C) 2005-2010 by Raymond Huang
 *   plushuang at users.sourceforge.net
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 *  ---
 *
 *  In addition, as a special exception, the copyright holders give
 *  permission to link the code of portions of this program with the
 *  OpenSSL library under certain conditions as described in each
 *  individual source file, and distribute linked combinations
 *  including the two.
 *  You must obey the GNU Lesser General Public License in all respects
 *  for all of the code used other than OpenSSL.  If you modify
 *  file(s) with this exception, you may extend this exception to your
 *  version of the file(s), but you are not obligated to do so.  If you
 *  do not wish to do so, delete this exception statement from your
 *  version.  If you delete this exception statement from all source
 *  files in the program, then also delete it here.
 *
 */

#include <string.h>

#include <glib.h>
#include <gmodule.h>

#include <ug_registry.h>


static GHashTable*	ug_hash_str	= NULL;
static GHashTable*	ug_hash_int	= NULL;


void		ug_registry_insert (const gchar* key_name, const gchar* key_type, gpointer value)
{
	gchar*		key;

	if (ug_hash_str == NULL)
		ug_hash_str = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);

	key = g_strconcat (key_name, "-", key_type, NULL);
	g_hash_table_insert (ug_hash_str, key, value);
//	key will be freed by g_hash_table_insert() in this case.
}

void		ug_registry_remove (const gchar* key_name, const gchar* key_type)
{
	gchar*		key;

	if (ug_hash_str == NULL)
		return;

	key = g_strconcat (key_name, "-", key_type, NULL);
	g_hash_table_remove (ug_hash_str, key);
	g_free (key);
}

gpointer	ug_registry_search (const gchar* key_name, const gchar* key_type)
{
	gchar*		key;
	gpointer	value;

	if (ug_hash_str == NULL)
		return NULL;

	key = g_strconcat (key_name, "-", key_type, NULL);
	value = g_hash_table_lookup (ug_hash_str, key);
	g_free (key);
	return value;
}


void		ug_registry_insert_int (gint key, gpointer value)
{
	if (ug_hash_int == NULL)
		ug_hash_int = g_hash_table_new (g_direct_hash, g_direct_equal);

	g_hash_table_insert (ug_hash_int, GINT_TO_POINTER (key), value);
}

void		ug_registry_remove_int (gint key)
{
	if (ug_hash_int == NULL)
		return;

	g_hash_table_remove (ug_hash_int, GINT_TO_POINTER (key));
}

gpointer	ug_registry_search_int (gint key)
{
	if (ug_hash_int == NULL)
		return NULL;

	return g_hash_table_lookup (ug_hash_int, GINT_TO_POINTER (key));
}




// ---------------------------------------------------------------------------
// register from file & folder
/*

#ifdef _WIN32
#define FILE_EXT        ".dll"
#define FILE_EXT_LEN    4
#else
#define FILE_EXT        ".so"
#define FILE_EXT_LEN    3
#endif

gboolean	ug_module_register_file	(const gchar* file_path)
{
	UgDataClass*	data_class;
	UgPluginClass*	plugin_class;
	UgOption*		option;
	GModule*		module;
	UgModuleInfo*	module_info;
	gboolean		result = FALSE;

	module = g_module_open (file_path, G_MODULE_BIND_LAZY);
	if (module == NULL) {
		g_message ("g_module_open() fail : %s.\n", g_module_error ());
		return FALSE;
	}

	if ( g_module_symbol (module, "ug_module_info", (gpointer)&module_info) ) {
		for (; obj_class; obj_class++)
			ug_class_register (obj_class);
		result = TRUE;
	}
//	g_module_close (module);

	return result;
}

guint	ug_module_register_folder	(const gchar* folder)
{
	const gchar*	filename;
	gchar*			full_path;
	guint			length;
	guint			count = 0;
	GDir*			dir;

	dir = g_dir_open (folder, 0, NULL);

	if (dir == NULL) {
		g_message ("ug_module_register_folder () : %s open fail\n", folder);
		return 0;
	}

	for (;;) {
		filename = g_dir_read_name (dir);
		if (filename == NULL)
			break;

		length = strlen (filename);
		if (length < FILE_EXT_LEN)
			continue;
		if (strncmp (filename +length -FILE_EXT_LEN, FILE_EXT, FILE_EXT_LEN) != 0)
			continue;

		full_path = g_build_filename (folder, filename, NULL);
		g_message ("ug_module_register_folder () : load %s\n", full_path);
		if (ug_module_register_file (full_path))
			count++;
	}

	return count;
}

*/
